@@ -83,6 +83,10 @@
 #include "mcs-channel.h"
 #include "mcs-manager.h"
 
+#ifndef PATH_MAX
+#define PATH_MAX 1024
+#endif
+
 struct _McsManager
 {
     Display *display;
@@ -801,177 +805,273 @@ static GMarkupParser markup_parser = {
     error_handler
 };
 
-McsChannel *
-mcs_manager_add_channel_from_file(McsManager *manager,
-        const gchar *channel_name, const gchar * filename)
+McsChannel*
+mcs_manager_add_channel_from_file (McsManager  *manager,
+                                   const gchar *channel_name,
+                                   const gchar *filename)
 {
-    gchar *contents;
-    GError *error;
-    GMarkupParseContext *context;
-    McsOptionParser parser;
-    McsChannel *channel;
-    struct stat sb;
+  gchar *contents;
+  GError *error;
+  GMarkupParseContext *context;
+  McsOptionParser parser;
+  McsChannel *channel;
+  struct stat sb;
 #ifdef HAVE_MMAP
-    void *addr;
+  void *addr;
 #endif
-    int fd;
+  size_t bytes;
+  int fd, rc;
 
-    g_return_val_if_fail(manager != NULL, NULL);
-    g_return_val_if_fail((filename != NULL) || (strlen(filename) > 0), NULL);
-    g_return_val_if_fail((channel_name != NULL) || (strlen(channel_name) > 0),
-        NULL);
+  g_return_val_if_fail (manager != NULL, NULL);
+  g_return_val_if_fail (filename != NULL && strlen (filename) > 0, NULL);
+  g_return_val_if_fail (channel_name != NULL && strlen (channel_name) > 0, NULL);
 
-    if ((channel = mcs_manager_add_channel(manager, channel_name)) == NULL)
-        return(NULL);
+  channel = mcs_manager_add_channel (manager, channel_name);
+  if (channel == NULL)
+    {
+      g_critical ("Unable to add channel \"%s\" to MCS manager",
+		  channel_name);
+      return NULL;
+    }
     
-#ifdef O_SHLOCK
-    if ((fd = open(filename, O_RDONLY | O_SHLOCK, 0)) < 0)
-#else
-    if ((fd = open(filename, O_RDONLY, 0)) < 0)
+  if (stat (filename, &sb) < 0)
+    {
+#if 0
+      /* The file doesn't exists, so what, it's not "critical"! */
+      g_critical ("Unable to stat file %s to load data of channel \"%s\": %s",
+		  filename,
+                  channel_name,
+		  g_strerror (errno));
 #endif
-        goto finished;
+      return NULL;
+    }
 
-    if (fstat(fd, &sb) < 0)
-        goto finished;
+  if ((fd = open (filename, O_RDONLY, 0)) < 0)
+    {
+      g_critical ("Unable to open file %s to load data of channel \"%s\": %s",
+		  filename,
+		  channel_name,
+		  g_strerror (errno));
+      return NULL;
+    }
 
 #ifdef HAVE_MMAP
-	/* Try to mmap(2) the config file, as this save us a lot of
-	 * kernelspace -> userspace copying
-	 */
+  /* Try to mmap(2) the config file, as this save us a lot of
+   * kernelspace -> userspace copying
+   */
 #ifdef MAP_FILE
-	addr = mmap(NULL, sb.st_size, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
+  addr = mmap (NULL, sb.st_size, PROT_READ, MAP_FILE | MAP_PRIVATE, fd, 0);
 #else
-	addr = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
+  addr = mmap (NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
 #endif
 
-	if (addr != NULL) {
-		/* nice, mmap did the job */
-		contents = addr;
-	}
-	else {
+  if (addr != NULL)
+    {
+      /* nice, mmap did the job */
+      contents = addr;
+    }
+  else
+    {
+      g_warning ("Failed to mmap file %s to load data of channel \"%s\": %s. "
+                 "Using read fallback.",
+                 filename,
+                 channel_name,
+                 g_strerror (errno));
 #endif
-		if ((contents = malloc((size_t)sb.st_size)) == NULL)
-			goto finished;
+      contents = g_malloc ((size_t)sb.st_size);
+      if (contents == NULL)
+        {
+          g_critical ("Unable to allocate %lu bytes of memory to load contents "
+                      "of file %s for channel \"%s\": %s",
+                      (unsigned long)sb.st_size,
+                      filename,
+                      channel_name,
+                      g_strerror (errno));
+          goto finished;
+        }
 
-		/* read(2):
-		 * --------
-		 * The system guarantees to read the number of bytes requested
-		 * if the descriptor references a normal file that has that 
-		 * many bytes left before the end-of-file.
-		 */
-		if (read(fd, contents, sb.st_size) < sb.st_size)
-			goto finished2;
+      for (bytes = 0; bytes < (size_t)sb.st_size; )
+        {
+          errno = 0;
+          rc = read (fd, contents + bytes, sb.st_size - bytes);
+          
+          if (rc < 0)
+            {
+              if (errno == EINTR || errno == EAGAIN)
+                continue;
+              
+              g_critical ("Unable to read contents from file %s: %s",
+                          filename,
+                          g_strerror (errno));
+              goto finished2;
+            }
+          else if (rc == 0)
+            {
+              g_critical ("Unexpected end of file reading contents from "
+                          "file %s: %s",
+                          filename,
+                          g_strerror (errno));
+            }
+          
+          bytes += rc;
+        }
+      
 #ifdef HAVE_MMAP
-	}
+    }
 #endif
 
-    error = NULL;
+  error = NULL;
 
-    parser.state = START;
-    parser.filename = filename;
-    parser.manager = manager;
-    parser.channel_name = channel_name;
+  parser.state = START;
+  parser.filename = filename;
+  parser.manager = manager;
+  parser.channel_name = channel_name;
 
-    context = g_markup_parse_context_new(&markup_parser, 0, &parser, NULL);
+  context = g_markup_parse_context_new (&markup_parser, 0, &parser, NULL);
 
-    if(!g_markup_parse_context_parse(context, contents, sb.st_size, NULL))
-        goto finished3;
+  if (!g_markup_parse_context_parse (context, contents, sb.st_size, &error))
+    {
+      g_critical ("Unable to parse file %s into channel \"%s\": %s",
+		  filename,
+		  channel_name,
+		  error->message);
+      g_error_free (error);
+      goto finished3;
+    }
 
-    if(!g_markup_parse_context_end_parse(context, NULL))
-        goto finished3;
+  if (!g_markup_parse_context_end_parse (context, &error))
+    {
+      g_critical ("Unable to parse file %s into channel \"%s\": %s",
+		  filename,
+		  channel_name,
+		  error->message);
+      g_error_free (error);
+      goto finished3;
+    }
 
-    mcs_manager_notify(manager, channel_name);
+  mcs_manager_notify (manager, channel_name);
 
-finished3:
-    g_markup_parse_context_free(context);
+ finished3:
+  g_markup_parse_context_free (context);
 
-finished2:
+ finished2:
 #ifdef HAVE_MMAP
-    if (addr != NULL)
-        munmap(addr, sb.st_size);
-    else
+  if (addr != NULL)
+    {
+      if (munmap (addr, sb.st_size) < 0)
+        {
+          g_critical ("Unable to unmap file %s with contents for channel "
+                      "\"%s\": %s. This should not happen!",
+                      filename,
+                      channel_name,
+                      g_strerror (errno));
+        }
+    }
+  else
 #endif
-        free(contents);
+    g_free (contents);
 
-finished:
-    (void)close(fd);
-    return(channel);
+ finished:
+  if (close (fd) < 0)
+    {
+      g_critical ("Failed to close file %s: %s",
+		  filename,
+		  g_strerror (errno));
+    }
+
+  return channel;
 }
 
 gboolean
-mcs_manager_save_channel_to_file(McsManager *manager, const gchar *channel_name,
-        const gchar *filename)
+mcs_manager_save_channel_to_file (McsManager  *manager,
+                                  const gchar *channel_name,
+                                  const gchar *filename)
 {
-    McsSetting *setting;
-    McsList *iter;
-    McsList *list;
-    FILE *fp;
-    int fd;
+  McsSetting *setting;
+  McsList *iter;
+  McsList *list;
+  FILE *fp;
+  gchar tmp_path[PATH_MAX];
     
-    g_return_val_if_fail(manager != NULL, FALSE);
-    g_return_val_if_fail((filename != NULL) || (strlen(filename) > 0), FALSE);
-    g_return_val_if_fail((channel_name != NULL) || (strlen(channel_name) > 0),
-            FALSE);
+  g_return_val_if_fail (manager != NULL, FALSE);
+  g_return_val_if_fail (filename != NULL
+			|| (strlen(filename) > 0), FALSE);
+  g_return_val_if_fail (channel_name != NULL
+			|| (strlen(channel_name) > 0), FALSE);
 
-#ifdef O_EXLOCK
-    if ((fd = open(filename, O_CREAT|O_EXLOCK|O_TRUNC|O_WRONLY, S_IRWXU)) < 0) {
-#else
-    if ((fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, S_IRWXU)) < 0) {
-#endif
-        g_warning(_("Cannot save file %s: %s"), filename, g_strerror(errno));
-	return(FALSE);
+  g_snprintf (tmp_path, PATH_MAX, "%s.tmp", filename);
+
+  fp = fopen (tmp_path, "w");
+  if (fp == NULL)
+    {
+      g_critical ("Unable to open file %s to store channel \"%s\" to: %s",
+		  tmp_path,
+		  channel_name,
+		  g_strerror (errno));
+      return FALSE;
     }
 
-    if ((fp = fdopen(fd, "w")) == NULL) {
-        (void)close(fd);
-        return(FALSE);
-    }
-  
-    /* Write header */
-    fprintf (fp,
-            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
-            "<!DOCTYPE mcs-option SYSTEM \"mcs-option.dtd\">\n"
-            "\n"
-            "<mcs-option>\n");
+  /* Write header */
+  fprintf (fp,
+	   "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+	   "<!DOCTYPE mcs-option SYSTEM \"mcs-option.dtd\">\n"
+	   "\n"
+	   "<mcs-option>\n");
 
-    list = mcs_manager_list_lookup(manager, channel_name);
-    
-	for (iter = list; iter != NULL; iter = iter->next) {
-	    setting = iter->setting;
+  list = mcs_manager_list_lookup (manager, channel_name);
+
+  for (iter = list; iter != NULL; iter = iter->next)
+    {
+      setting = iter->setting;
            
-	    switch (setting->type) {
-		case MCS_TYPE_INT:
-            fprintf (fp,
-                    "\t<option name=\"%s\" type=\"int\" value=\"%i\"/>\n",
-                    setting->name, setting->data.v_int);
-		    break;
+      switch (setting->type) {
+      case MCS_TYPE_INT:
+	fprintf (fp,
+		 "\t<option name=\"%s\" type=\"int\" value=\"%i\"/>\n",
+		 setting->name, setting->data.v_int);
+	break;
 		
-        case MCS_TYPE_COLOR:
-            fprintf (fp, "\t<option name=\"%s\" type=\"color\" "
-                    "value=\"%16u,%16u,%16u,%16u\"/>\n",
-                    setting->name,
-                    setting->data.v_color.red,
-                    setting->data.v_color.green,
-                    setting->data.v_color.blue,
-                    setting->data.v_color.alpha);
-		    break;
+      case MCS_TYPE_COLOR:
+	fprintf (fp, "\t<option name=\"%s\" type=\"color\" "
+		 "value=\"%16u,%16u,%16u,%16u\"/>\n",
+		 setting->name,
+		 setting->data.v_color.red,
+		 setting->data.v_color.green,
+		 setting->data.v_color.blue,
+		 setting->data.v_color.alpha);
+	break;
 	        
-        case MCS_TYPE_STRING:
-            fprintf (fp, "\t<option name=\"%s\" type=\"string\" "
-                    "value=\"%s\"/>\n",
-                    setting->name,
-                    setting->data.v_string);
-		    break;
+      case MCS_TYPE_STRING:
+	fprintf (fp, "\t<option name=\"%s\" type=\"string\" "
+		 "value=\"%s\"/>\n",
+		 setting->name,
+		 setting->data.v_string);
+	break;
 
-        default:
-            break;
-	    }
+      default:
+	break;
+      }
     }
 
-    fprintf (fp, "</mcs-option>\n");
-    
-    fflush(fp);
-    fclose(fp);
+  fprintf (fp, "</mcs-option>\n");
 
-    return(TRUE);
+  if (fclose (fp) == EOF)
+    {
+      g_critical ("Unable to close file handle for %s: %s",
+		  tmp_path,
+		  g_strerror (errno));
+      unlink (tmp_path);
+      return FALSE;
+    }
+
+  if (rename (tmp_path, filename) < 0)
+    {
+      g_critical ("Unable to rename file %s to %s: %s",
+		  tmp_path,
+		  filename,
+		  g_strerror (errno));
+      return FALSE;
+    }
+
+  return TRUE;
 }
